#include "aeon.h"

static char logfile[MAX_LEN]; 			/* path to user logfile */
	
int main(int argc, char **argv)
{
	char settings[MAX_SETTINGS][MAX_LEN];	/* non-parsed cfg lines			*/
	char sender[PATH_LEN+1], smtpServer[DOM_LEN+1], myHostname[DOM_LEN+1];
	char login[DOM_LEN+1], password[DOM_LEN+1]; 
	char *addresses[MAX_RCPT]; 		/* recipients 				*/
	char headerFile[] = "/tmp/aeonXXXXXX";	/* temp file with header lines		*/
	int numSet;				/* number of cfg lines in rc file	*/
	int numRcpt;				/* number or recipients 		*/
	int sd; 				/* socket descriptor 			*/
	int authMethod = 0;			/* authentication method 		*/
	int i; 					/* needed iterator 			*/

	numSet = getConfig(settings);		/* check for rc file and get the lines	*/
	if (numSet == -1) {
		logEntry("Missing config file: ~/.aeonrc! ABORTING!\n");
		exit(1);
	}
	switch (setup(settings, numSet, sender, smtpServer, myHostname, login, password, logfile)) {
		case -1: /* very, very bad... */
			logTimestamp(logfile);
			logEntry("Missing important settings (sender, server)! ABORTING!\n");
			exit(1);
		case 1: /* SMTP "without authentication" mode */
			logTimestamp(logfile);
			logEntry("Could not find login/pass. ");
			logEntry("Aeon will try to send this message without authentication.\n");
			authMethod = AUTH_NONE;
			break;
		case 0: /* just all what we need */
			logTimestamp(logfile);
			logEntry("Config seems to be OK. Getting to work.\n");
			break;
	}
	
	/* get recipients list and subject line from message header */
	numRcpt = getMessageHeader(addresses, headerFile);
	switch (numRcpt) {
		case -1: /* mkstemp failed. exiting */
			logEntry("Could not create temporary header file in /tmp. ABORTING!\n");
			exit(1);
		case 0: /* email addresses missing. exiting */
			logEntry("Improper program call. No recipients. ABORTING!\n");
			exit(1);
		default:
			;
	}

	sd = establishConnection(smtpServer); 	/* open socket, receive greets */ 
	if (sd < 0) { 				/* something wrong with socket preparation */
		logEntry("Could not prepare connection, aborting...\n");
		for (i=0; i<numRcpt; i++) free(addresses[i]);
		exit(1);
	}
	if (sd == 0) { 				/* something wrong with server */
		logEntry("Could not connect to server, aborting...\n");
		for (i=0; i<numRcpt; i++) free(addresses[i]);
		exit(1);
	}
	
	if (authMethod != AUTH_NONE) authMethod = sayEHLO(sd, myHostname); /* EHLO hostname */ 
	switch (authMethod) { 
		case AUTH_PLAIN:
			logEntry("Found a way to authenticate - PLAIN\n");
			if (authorizePlain(sd, login, password) != 0) 
				logEntry("Authentication failed, check user/pass.\n");
			break;
		case AUTH_NONE:
			logEntry("Authentication disabled or no suitable protocol found. ");
			logEntry("Sending without authentication!\n");
			break;
		case -1:
			logEntry("Error while greeting the server!\n");
			for (i=0; i<numRcpt; i++) free(addresses[i]);
			exit(1);
	}
	
	if (sayWho(sd, sender) != 0) { 			/* MAIL FROM:<sender@oddress.com> */
		logEntry("Something wrong with sender address, check config.\n");
		for (i=0; i<numRcpt; i++) free(addresses[i]);
		exit(1);
	} 

	for (i=0; i<numRcpt; i++) 			/* send to all recipients */
		if (sayWhere(sd, addresses[i]) != 0) { 	/* RCPT TO:<addresses[i]> */
			logEntry("Wrong recipient: ");
			logEntry(addresses[i]);
			logEntry("\n");
		}
	if (sendMessageHeader(sd, headerFile) != 0) { 
		logEntry("Could not send the message's header, aborting...\n");
		for (i=0; i<numRcpt; i++) free(addresses[i]);
		exit(1);
	}
        if (sendData(sd) != 0) { 			/* DATA, message, \r\n.\r\n */
		logEntry("Could not send the message, aborting...\n");
		for (i=0; i<numRcpt; i++) free(addresses[i]);
		exit(1);
	} 
	if (sayQuit(sd) != 0) 				/* QUIT, close the socket */
		logEntry("Mail sent, but could not clean up or say goodbye to the server!\n");

	for (i=0; i<numRcpt; i++) free(addresses[i]);
	
	return 0;
}

int establishConnection(char *serverName)
{
	int mySocket;
	struct sockaddr_in addr;
	struct hostent hisData, *hisDataPtr;
	char resp[MAX_LEN];

	bzero(resp, MAX_LEN);
	mySocket = socket(AF_INET, SOCK_STREAM, 0);	/* creating socket */
	if (mySocket < 0) {
		perror("socket");
		return -1;
	}
	
	/* filling sockaddr_in addr structure... */
	hisDataPtr = &hisData;
	if ((hisDataPtr = gethostbyname(serverName)) == NULL) {
		perror("gethostbyname");
		return -1;
	}
	bcopy((char*) hisDataPtr->h_addr, (char*) &addr.sin_addr.s_addr,
				          hisDataPtr->h_length);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(PORT);
	/* and connecting */
	if (connect(mySocket, (struct sockaddr*) &addr, sizeof(addr)) == -1) {
		perror("connect");
		return -1;
	}
	/* waiting for server's 220 code */
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) {
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_READY) {
		/* socket ready, but server failed, aborting... */
		close(mySocket);
		logEntry(resp);
		return 0;
	}
	logEntry("Connecting to ");
	logEntry(serverName);
	logEntry(" OK!\n");

	return mySocket;
}

int sayEHLO(int mySocket, char *myHostname)
{
	char resp[MAX_LEN];
	char ehlo[DOM_LEN+8] = "EHLO ";

	bzero(resp, MAX_LEN);
	strcat(ehlo, myHostname);
	strcat(ehlo, "\r\n"); 			/* creating "EHLO hostname" string */
	
	if (send(mySocket, (void*) ehlo, strlen(ehlo), 0) == -1) {
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) {
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_ALLOK) {
		logEntry(resp);
		return -1;
	}

	if (strstr(resp, "PLAIN")) return AUTH_PLAIN;
	else return AUTH_NONE;
}

int authorizePlain(int mySocket, char *login, char *password)
{
	char resp[MAX_LEN];
	char authPlain[3*DOM_LEN+3]; 			/* login\0login\0password */
	char authB64[4*DOM_LEN];			/* authPlain base64 encoded */
	char authLine[MAX_LEN] = "AUTH PLAIN ";
	int len;

	bzero(resp, MAX_LEN);
	len = 2*strlen(login)+strlen(password)+2;	
	authCat(login, password, authPlain);		/* concat: login\0login\0pass */
	encodeB64(authPlain, len, authB64, 4*DOM_LEN);	/* encode it with base64 */
	strcat(authLine, authB64);			/* join the pieces  */
	strcat(authLine, "\r\n");	

	if (send(mySocket, (void*) authLine, strlen(authLine), 0) == -1) {
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) {
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_AUTOK) {
		logEntry(resp);
		return 1;
	}
	
	return 0;
}

int sayWho(int mySocket, char *mailAddress)
{
	char resp[MAX_LEN];
	char who[PATH_LEN+15] = "MAIL FROM:<";

	bzero(resp, MAX_LEN);
	strcat(who, mailAddress);
	strcat(who, ">\r\n"); 			/* create "MAIL FROM:<some@body.com>" string */
	
	if (send(mySocket, (void*) who, strlen(who), 0) == -1) {
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) {
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_ALLOK) {
		logEntry(resp);
		return 1;
	}

	return 0;
}

int sayWhere(int mySocket, char *mailAddress)
{
	char resp[MAX_LEN];
	char where[PATH_LEN+13] = "RCPT TO:<";

	bzero(resp, MAX_LEN);
	strcat(where, mailAddress);
	strcat(where, ">\r\n"); 		/* create "RCPT TO:<some@body.com" string */
	
	if (send(mySocket, (void*) where, strlen(where), 0) == -1) {
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) {
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_ALLOK) {
		logEntry(resp);
		return 1;
	}

	return 0;
}

/* send DATA command and previously saved message header line by line */
int sendMessageHeader(int mySocket, char *headerFile)
{
	char resp[MAX_LEN];
	char data[] = "DATA\r\n";
	FILE *header;
	char headerLine[TXT_LEN];
	
	bzero(resp, MAX_LEN);

	if (send(mySocket, data, strlen(data), 0) == -1) { 	/* send DATA cmd */
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) { 	/* receive permission */
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_SMAIL) {
		logEntry(resp);
		return 1;
	}
	
	if ((header = fopen(headerFile, "r")) == NULL) {
		perror("fopen");
		return -1;
	}
	while (fgets(headerLine, TXT_LEN-3, header)) {
		headerLine[strlen(headerLine)-1] = '\0';	/* remove bare <lf> */
		strcat(headerLine, "\r\n");
		if (send(mySocket, (void*) headerLine, strlen(headerLine), 0) == -1) { 
			perror("send");
			return -1;
		}
	}
	fclose(header);
	if (unlink(headerFile) == -1) {
		logEntry("Warning! Could not remove temporary header file: ");
		logEntry(headerFile);
		logEntry("\n");
	}
	
	
	return 0;
}

int sendData(int mySocket)
{
	char resp[MAX_LEN];
	char msgLine[TXT_LEN];
	
	
	while (fgets(msgLine, TXT_LEN-3, stdin) != NULL) { 	/* start sending msg lines */
		msgLine[strlen(msgLine)-1] = '\0'; 		/* remove bare <lf> */
		strcat(msgLine, "\r\n");
		if (send(mySocket, (void*) msgLine, strlen(msgLine), 0) == -1) {
			perror("send");
			return -1;
		}
	}
	if (send(mySocket, (void*) "\r\n.\r\n", 5, 0) == -1) { 	/* end of msg */
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) { 	/*recv OK */
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_ALLOK) {
		logEntry(resp);
		return 1;
	}
	else logEntry("DONE. Sending mail OK.\n");

	return 0;
}
	
int sayQuit(int mySocket)
{
	char quit[] = "QUIT\r\n";
	char resp[MAX_LEN];
	
	bzero(resp, MAX_LEN);

	if (send(mySocket, (void*) quit, strlen(quit), 0) == -1) { 	/* say bye */
		perror("send");
		return -1;
	}
	if (recv(mySocket, (void*) resp, MAX_LEN, 0) == -1) { 		/*rcv bye */
		perror("recv");
		return -1;
	}
	if (parseReply(resp) != REP_CLOSE) { 				/* smth wrong? */
		logEntry(resp);
		return 1; 
	}
	if (close(mySocket) == -1) {
		perror("close");
		return -1;
	}
	
	return 0;
}

/* write msg into logfile or stderr */
int logEntry(const char *entry)
{
	FILE *lf; 				/* log file handler */

	lf = fopen(logfile, "a");
	if (lf) {
		fprintf(lf, "%s", entry);
		fclose(lf);
	}
	else fprintf(stderr, "%s", entry);

	return 0;
}
