/* This file is copied from GNU libc - libc/resolv/base64.c */

/*
 * Copyright (c) 1996-1999 by Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
 * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
 * Portions Copyright (c) 1995 by International Business Machines, Inc.
 *
 * International Business Machines, Inc. (hereinafter called IBM) grants
 * permission under its copyrights to use, copy, modify, and distribute this
 * Software with or without fee, provided that the above copyright notice and
 * all paragraphs of this notice appear in all copies, and that the name of IBM
 * not be used in connection with the marketing of any product incorporating
 * the Software or modifications thereof, without specific, written prior
 * permission.
 *
 * To the extent it has a right to do so, IBM grants an immunity from suit
 * under its patents, if any, for the use, sale or manufacture of products to
 * the extent that such products are used for performing Domain Name System
 * dynamic updates in TCP/IP networks by means of the Software.  No immunity is
 * granted for any product per se or for any other function of any product.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", AND IBM DISCLAIMS ALL WARRANTIES,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.  IN NO EVENT SHALL IBM BE LIABLE FOR ANY SPECIAL,
 * DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE, EVEN
 * IF IBM IS APPRISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#include "aeon.h"
#include <stddef.h>		/* size_t */
#include <ctype.h>

#define Assert(Cond) if (!(Cond)) abort()

static const char Base64[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char Pad64 = '=';


/**
 * gsasl_base64_encode:
 * @src: input byte array
 * @srcLen: size of input byte array
 * @targ: output byte array
 * @targLen: size of output byte array
 * 
 * Encode data as base64.  Converts characters, three at a time,
 * starting at src into four base64 characters in the targ area
 * until the entire input buffer is encoded.
 * 
 * Return value: Returns the number of data bytes stored at the
 * targ, or -1 on error.
 **/
int encodeB64(char const *src, size_t srcLen, char *targ, size_t targLen)
{
	size_t datalength = 0;
	unsigned char input[3];
	unsigned char output[4];
	size_t i;

	while (2 < srcLen) {
		input[0] = *src++;
		input[1] = *src++;
	        input[2] = *src++;
  	        srcLen -= 3;

	        output[0] = input[0] >> 2;
  		output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
 	        output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
	        output[3] = input[2] & 0x3f;
	        Assert (output[0] < 64);
 	        Assert (output[1] < 64);
	        Assert (output[2] < 64);
	        Assert (output[3] < 64);

	        if (datalength+4 > targLen) return (-1);
	        targ[datalength++] = Base64[output[0]];
	        targ[datalength++] = Base64[output[1]];
 	        targ[datalength++] = Base64[output[2]];
	        targ[datalength++] = Base64[output[3]];
	}

	/* Now we worry about padding. */
  	if (0 != srcLen) {
      		/* Get what's left. */
		input[0] = input[1] = input[2] = '\0';
	        for (i = 0; i < srcLen; i++) input[i] = *src++;

	        output[0] = input[0] >> 2;
	        output[1] = ((input[0] & 0x03) << 4) + (input[1] >> 4);
 	        output[2] = ((input[1] & 0x0f) << 2) + (input[2] >> 6);
	        Assert (output[0] < 64);
	        Assert (output[1] < 64);
	        Assert (output[2] < 64);

	        if (datalength+4 > targLen) return (-1);
	        targ[datalength++] = Base64[output[0]];
	        targ[datalength++] = Base64[output[1]];
	        if (srcLen == 1) targ[datalength++] = Pad64;
	        else targ[datalength++] = Base64[output[2]];
	        targ[datalength++] = Pad64;
    	}
       	if (datalength >= targLen) return (-1);
	targ[datalength] = '\0';	/* Returned value doesn't count \0. */
	
	return (datalength);
}

